#pragma once

#include <cstddef>
#include <synchapi.h>
#include "storm.h"

template<typename T>
class ThreadSafeMemoryChain {
public:

	struct Node {
		Node* volatile next;
	};

	T* volatile first;

	void push(T* item, size_t last_item_offset) volatile {
		T* current_top;
		bool swapped;

		do {
			current_top = first;
			((T*)((char*)item + last_item_offset))->next = current_top;
			T* previous_top = (T*)InterlockedCompareExchange(
				reinterpret_cast<volatile LONG*>(&first),
				reinterpret_cast<LONG>(item),
				reinterpret_cast<LONG>(current_top)
			);
			swapped = previous_top == current_top;
		} while (!swapped);
	}

	T* pop() volatile {
		while (true) {
			T* current_top = first;
			if (!current_top) {
				break;
			}

			T* previous_top = (T*)InterlockedCompareExchange(
				reinterpret_cast<volatile LONG*>(&first),
				reinterpret_cast<LONG>(current_top->next),
				reinterpret_cast<LONG>(current_top)
			);
			if (previous_top == current_top) {
				return current_top;
			}
		}

		return nullptr;
	}
};

class CDataRecycler {
public:

	struct _Node: ThreadSafeMemoryChain<_Node>::Node {
		void* buffer;
		size_t size;
	};

	struct _Cluster: ThreadSafeMemoryChain<_Cluster>::Node {
		CDataRecycler::_Node nodes[];
	};

	struct _VMT {
		int (__thiscall *method00)(CDataRecycler* self);
		int (__thiscall *method04)(CDataRecycler* self);
		void *(__thiscall *allocate)(CDataRecycler* self, size_t *size, size_t *allocated_memory_size, const char *source_file, int source_line);
		void *(__thiscall *reallocate)(CDataRecycler* self, void *memory, size_t size, size_t *allocated_memory_size, const char *source_file, int source_line);
		void (__thiscall *deallocate)(CDataRecycler* self, void* memory, const char* source_file, int source_line);
	};

	void* _vtable;
	volatile int _cache_slots_count;
	int _cluster_size;
	ThreadSafeMemoryChain<_Cluster> volatile _clusters;
	ThreadSafeMemoryChain<_Node> volatile _empty_nodes;
	ThreadSafeMemoryChain<_Node> volatile _nodes;

	_Node* _get_node();

	bool get_data(void*& buffer, size_t& size);
	void put_data(void* buffer, size_t size);
};

template<typename T>
class TInstanceRecycler: public CDataRecycler {
public:
	size_t instance_size;

	template<typename ...Args>
	T* get_item(Args... args) {
		void* buffer;
		size_t size;

		if (!get_data(buffer, size) || (size < sizeof(T))) {
			buffer = SMemRealloc(buffer, sizeof(T), __FILE__, __LINE__, 0);
		}

		return new (buffer) T(args...);
	}
};
