#pragma once

#include <cstddef>
#include <type_traits>
#include <windef.h>
#include "wc3/linkedlist.hpp"
#include "wc3/data_recycler.hpp"

class CDataStore {
public:

	struct VMT {
		void (__thiscall* init_buffer)(CDataStore* self, int a1, int a2, int a3);
		void (__thiscall* free_buffer)(CDataStore* self, void** data, size_t* offset, size_t* capacity);
		BOOL (__thiscall *prepare_to_read)(CDataStore* self, size_t position, size_t count, void** p_data, size_t* p_offset, size_t* p_capacity);
		BOOL (__thiscall* prepare_to_write)(CDataStore* self, size_t position, size_t count, void** p_data, size_t* p_offset, size_t* p_capacity, const char* source_file, int source_line);
		void (__thiscall* operator_delete)(CDataStore* self, bool free_memory);
		int (__thiscall* method14)(CDataStore* self);
		BOOL (__thiscall* is_out_of_bounds)(CDataStore* self);
		void (__thiscall* clear)(CDataStore* self);
		void (__thiscall* reset_position)(CDataStore* self);
		void (__thiscall* get_buffer)(CDataStore* self, void** data, size_t* size, size_t* capacity);
		int (__thiscall* pop_buffer)(CDataStore* self, void** data, size_t* size, size_t* capacity);
	};

	void* vtable;
	void* data;
	size_t current_region_start;
	size_t capacity;
	size_t size;
	size_t position;

	bool _request_region(size_t region_position, size_t region_size);

	void clear();
	bool write(const void* buffer, size_t count);
};

template<typename T>
CDataStore& operator << (CDataStore& store, const T& value) {
	static_assert(std::is_arithmetic_v<T>, "default implementation only for simple types");
	store.write(&value, sizeof(T));
	return store;
}

class CDataStoreChunked: public CDataStore {
public:

	struct VMT: CDataStore::VMT {
		int (__thiscall *method2C)(CDataStoreChunked* self);
		int (__thiscall *method30)(CDataStoreChunked* self);
		void (__thiscall *free_chunk)(CDataStoreChunked* self, void *chunk, DWORD chunk_size);
		int (__thiscall *method38)(CDataStoreChunked* self);
	};

	struct Chunk {};

	TInstanceRecycler<Chunk> recycler;
	int flags;
	size_t chunk_size;
	void* current_chunk;
	WORD* chunk_double_sized;
	int field44;
	int field48_usedchunks_;
	int field4C_pchunk;
	TSLinkedList<void*> chunks__;
	int field5C_freedchunks___;
	int field60;
	int field64;
	int field68;
};

template<size_t cache_size>
class CDataStoreCache: public CDataStore {
public:
	unsigned char cache[cache_size];
};
