#pragma once

#include "wc3/geometry.hpp"
#include "wc3/array.hpp"
#include "wc3/linkedlist.hpp"
#include "wc3/observer.hpp"
#include "wc3/instance_table.hpp"

struct EVENT_DATA_MOUSE {
	int mouse_locked_;
	int mouse_key;
	int mouse_keys;
	int modifiers;
	int flags_;
	float x;
	float y;
	int wheel_delta;
	int time;
};

enum MOUSE_BUTTON_FLAGS
{
	LEFT_MOUSE_BUTTON = 0x1,
	MIDDLE_MOUSE_BUTTON = 0x2,
	RIGHT_MOUSE_BUTTON = 0x4,
	X1_MOUSE_BUTTON = 0x8,
	X2_MOUSE_BUTTON = 0x10,
};

class CMouseEvent: public CEvent, public EVENT_DATA_MOUSE {};

enum WAR3KEY {
	KEY_0 = 0x30,
	KEY_1 = 0x31,
	KEY_2 = 0x32,
	KEY_3 = 0x33,
	KEY_4 = 0x34,
	KEY_5 = 0x35,
	KEY_6 = 0x36,
	KEY_7 = 0x37,
	KEY_8 = 0x38,
	KEY_9 = 0x39,
	KEY_A = 0x41,
	KEY_B = 0x42,
	KEY_C = 0x43,
	KEY_D = 0x44,
	KEY_E = 0x45,
	KEY_F = 0x46,
	KEY_G = 0x47,
	KEY_H = 0x48,
	KEY_I = 0x49,
	KEY_J = 0x4A,
	KEY_K = 0x4B,
	KEY_L = 0x4C,
	KEY_M = 0x4D,
	KEY_N = 0x4E,
	KEY_O = 0x4F,
	KEY_P = 0x50,
	KEY_Q = 0x51,
	KEY_R = 0x52,
	KEY_S = 0x53,
	KEY_T = 0x54,
	KEY_U = 0x55,
	KEY_V = 0x56,
	KEY_W = 0x57,
	KEY_X = 0x58,
	KEY_Y = 0x59,
	KEY_Z = 0x5A,
	KEY_BACK = 0x202,
	KEY_TAB = 0x203,
	KEY_RETURN = 0x201,
	KEY_SHIFT = 0x0,
	KEY_CONTROL = 0x1,
	KEY_MENU = 0x2,
	KEY_PAUSE = 0x211,
	KEY_CAPITAL = 0x20E,
	KEY_ESCAPE = 0x200,
	KEY_SPACE = 0x20,
	KEY_PRIOR = 0x20C,
	KEY_NEXT = 0x20D,
	KEY_END = 0x20B,
	KEY_HOME = 0x20A,
	KEY_LEFT = 0x204,
	KEY_RIGHT = 0x206,
	KEY_SNAPSHOT = 0x212,
	KEY_INSERT = 0x208,
	KEY_DELETE = 0x209,
	KEY_NUMPAD0 = 0x101,
	KEY_NUMPAD1 = 0x102,
	KEY_NUMPAD2 = 0x103,
	KEY_NUMPAD3 = 0x104,
	KEY_NUMPAD4 = 0x105,
	KEY_NUMPAD5 = 0x106,
	KEY_NUMPAD6 = 0x107,
	KEY_NUMPAD7 = 0x108,
	KEY_NUMPAD8 = 0x109,
	KEY_NUMPAD9 = 0x10A,
	KEY_MULTIPLY = 0x10D,
	KEY_ADD = 0x10B,
	KEY_SUBTRACT = 0x10C,
	KEY_DECIMAL = 0x10F,
	KEY_DIVIDE = 0x10E,
	KEY_NUMLOCK = 0x20F,
	KEY_SCROLL = 0x210,
	KEY_OEM_1 = 0x115,
	KEY_OEM_PLUS = 0x110,
	KEY_OEM_COMMA = 0x117,
	KEY_OEM_MINUS = 0x111,
	KEY_OEM_PERIOD = 0x118,
	KEY_OEM_2 = 0x119,
	KEY_OEM_3 = 0x100,
	KEY_OEM_4 = 0x112,
	KEY_OEM_5 = 0x114,
	KEY_OEM_6 = 0x113,
	KEY_OEM_7 = 0x116,
	KEY_F1 = 0x300,
	KEY_F2 = 0x301,
	KEY_F3 = 0x302,
	KEY_F4 = 0x303,
	KEY_F5 = 0x304,
	KEY_F6 = 0x305,
	KEY_F7 = 0x306,
	KEY_F8 = 0x307,
	KEY_F9 = 0x308,
	KEY_F10 = 0x309,
	KEY_F11 = 0x30A,
	KEY_F12 = 0x30B,
};

struct EVENT_DATA_KEY {
	int key;
	int modifiers;
	int lparam;
	int time;
};

class CKeyEvent: public CEvent, public EVENT_DATA_KEY {};

const int EVENT_TYPE_KEY_DOWN = 0x40060064;
const int EVENT_TYPE_MOUSE_DOWN = 0x400500C8;
const int EVENT_TYPE_MOUSE_UP = 0x400500C9;

enum class EVT_TYPE
{
	CHAR = 0x1,
	APPLICATION_ACTIVATED = 0x2,
	MESSAGES_BEGIN = 0x3,
	MESSAGES_END = 0x4,
	TICK = 0x5,
	SHUTDOWN = 0x6,
	INIT____ = 0x7,
	KEY_DOWN = 0x8,
	KEY_UP = 0x9,
	KEY_REPEAT = 0xA,
	MOUSE_DOWN = 0xB,
	MOUSE_UNK_C = 0xC,
	MOUSE_MOVE_DELTA = 0xD,
	MOUSE_UP = 0xE,
	MOUSE_MOVE = 0x10,
	MAP_CHECK_ = 0x12,
	NET_DATA_EVENT = 0x18,
	IME = 0x1B,
	WINDOW_SIZE_CHANGED = 0x1C,
};

class SomeEvtContextArray: public TSGrowableArray<void*> {
public:
	size_t metadata_offset_;
};

class EvtHandler: public LinkedNode<EvtHandler> {
public:
	typedef BOOL (__fastcall *Callback)(void *event_data, void *event_param);

	Callback callback;
	void* parameter;
	float order;
	BOOL fake;
};

class EvtContext: public TInstanceId<EvtContext> {
public:
	CRITICAL_SECTION pressed_keys_lock_andF2C;
	DWORD last_game_loop_update;
	BOOL shutdown_scheduled_;
	DWORD timer_priority_vmt;
	SomeEvtContextArray* array34;
	DWORD array34_index;
	DWORD field3C_last_hz_time_;
	DWORD creation_time_or_init_time_;
	DWORD flags1_1inited_2read_events_4tick_emited_;
	DWORD idleTime48;
	DWORD idleTime4C;
	DWORD flags2_50;
	DWORD flags2_54;
	DWORD flags2_50_54_comparison_result_;
	TSLinkedList<EvtHandler> event_handlers_registry[29];
	TSLinkedList<void*> field1B8;
	DWORD pressed_mouse_keys_;
	TSLinkedList<void*> pressed_keys;
	DWORD field1D4;
	DWORD field1D8;
	DWORD field1DC;
	DWORD field1E0;
	DWORD field1E4;
	DWORD field1E8;
	DWORD field1EC;
	DWORD field1F0;
	DWORD field1F4;
	DWORD field1F8;
	DWORD field1FC;
	DWORD field200;
	DWORD field204;
	DWORD wc3props;
	void* oscallL309;
	BOOL field210;
};


using EvtContextTable = TInstanceIdTable<EvtContext, 8>;
extern EvtContextTable* p_evt_context_table;

typedef void (__fastcall* emit_event_t)(EvtContext *evt_context, EVT_TYPE event_id, void *event_data);
extern emit_event_t emit_event;

typedef BOOL (__fastcall* EventCallback)(void* event_data, void* parameter);

typedef void (__fastcall* register_event_handler_t)(EVT_TYPE event_id, EventCallback callback, void *param, float order);
extern register_event_handler_t register_event_handler;

typedef void (__fastcall* unregister_event_handler_t)(EVT_TYPE event_id, EventCallback callback);
extern unregister_event_handler_t unregister_event_handler;

typedef void (__fastcall* TGetMousePosition)(float* out_x, float* out_y);
extern TGetMousePosition get_mouse_position;

extern HWND* p_game_window;

Vector2f get_mouse_position_percents();
