#pragma once

#include "wc3/class_utils.hpp"
#include "wc3/linkedlist.hpp"
#include "synchapi.h"

template<typename T>
class TInstanceId: public InsertVMT, public LinkedNode<T> {
public:
	struct VMT {
		void (__thiscall* operator_delete)(TInstanceId* self, bool free_memory);
	};

	int handle;
};

template<typename T>
class TRefInstanceId: public TInstanceId<T> {
public:
	struct VMT {
		void (__thiscall* release)(TRefInstanceId* self);
		const char* (__thiscall* get_string_path____)(TRefInstanceId* self);
	};
};

class TInstanceIdTableLock;

typedef void (__thiscall* TInstanceIdTableLockFunction)(TInstanceIdTableLock* self, BOOL unk);
extern TInstanceIdTableLockFunction TInstanceIdTableLock_acquire;
extern TInstanceIdTableLockFunction TInstanceIdTableLock_release;

class TInstanceIdTableLockSubObject {
public:
	int field00;
	int field04_hi11bitsIsIndex;
};

class TInstanceIdTableLock: public TInstanceIdTableLockSubObject {
public:
	int field08;

	void acquire(BOOL unk) {
		TInstanceIdTableLock_acquire(this, unk);
	}

	void release(BOOL unk) {
		TInstanceIdTableLock_release(this, unk);
	}
};

template<typename T, size_t branch_count>
class TInstanceIdTable {
public:
	CRITICAL_SECTION lock;
	int field04;
	int field08;
	TInstanceIdTableLock locks[branch_count];
	TSLinkedList<T> lists[branch_count];

	T* get_item(int handle) {
		int index = handle % branch_count;
		locks[index].acquire(false);

		// Don't use default iterator because lists contain the wrong offset value.
		T* current = lists[index]._root.value();
		while (current && current->handle != handle) {
			current = static_cast<LinkedNode<T>*>(current)->value();
		}

		if (current) {
			return current;
		}
		else {
			locks[index].release(false);
			return nullptr;
		}
	}

	void unlock_item(int handle) {
		locks[handle % branch_count].release(false);
	}
};

template<typename T, size_t branch_count>
class TRefInstanceIdTable: public InsertVMT, public TInstanceIdTable<T, branch_count> {
public:
	struct VMT {
		void (__thiscall* free_object_)(TRefInstanceIdTable* self, T* object);
	};
};
