#pragma once

#include <cstddef>
#include <cstdint>
#include <iterator>

template<typename T>
struct LinkedNode {
	LinkedNode* previous;
	T* _value;

	bool has_next() const {
		return reinterpret_cast<intptr_t>(_value) > 0;
	}

	T* value() const {
		return has_next() ? _value : nullptr; 
	}

	LinkedNode& next(size_t offset) {
		if (has_next()) {
			return *(LinkedNode*)((char*)_value + offset);
		}
		else {
			return *(LinkedNode*)~(uintptr_t)_value;
		}
	}

	LinkedNode& next() {
		if (has_next()) {
			T* value = previous->value();
			if (value) {
				size_t offset = (char*)this - (char*)value;
				return next(offset);
			}
			else {
				throw "Can't calculate offset for root node.";
			}
		}
		else {
			return *(LinkedNode*)~(uintptr_t)_value;
		}
	}

	void unlink() {
		previous->_value = _value;
		next().previous = previous;
	}

	void loop() {
		previous = this;
		_value = reinterpret_cast<T*>(~reinterpret_cast<uintptr_t>(this));
	}

	void clear() {
		_value = nullptr;
		previous = nullptr;
	}
};

template<typename T>
class TSLinkedList {
public:

	class Iterator {
	public:
		friend class TSLinkedList;

		LinkedNode<T>* node;

		Iterator(LinkedNode<T>* node) {
			this->node = node;
		}

	public:
		typedef std::bidirectional_iterator_tag iterator_category;
		typedef T                          value_type;
		typedef ptrdiff_t                  difference_type;
		typedef T*                         pointer;
		typedef T&                         reference;

		bool operator == (const Iterator& other) const {
			return node == other.node;
		}

		bool operator != (const Iterator& other) const {
			return !(*this == other);
		}

		operator T* () const {
			return node->previous->value();
		}

		T& operator * () const {
			return *node->previous->value();
		}

		T* operator -> () const {
			return node->previous->value();
		}

		Iterator& operator ++ () {
			node = &node->next();
			return *this;
		}

		Iterator operator ++ (int) {
			Iterator temp = *this;
			++(*this);
			return temp;
		}

		Iterator& operator -- () {
			node = node->previous;
			return *this;
		}

		Iterator operator -- (int) {
			Iterator temp = *this;
			--(*this);
			return temp;
		}
	};

	using iterator = Iterator;

	TSLinkedList(size_t node_offset) {
		_node_offset = node_offset;
		_root.loop();
	}

	TSLinkedList(TSLinkedList&& other) {
		steal_content(other);
	}

	~TSLinkedList() {
		_root.unlink();
	}

	TSLinkedList& operator = (TSLinkedList&& other) {
		_root.unlink();
		steal_content(other);
	}

	size_t _node_offset;
	LinkedNode<T> _root;

	auto begin() {
		LinkedNode<T>& node = _root.next(_node_offset);
		return Iterator(&node);
	}

	auto end() {
		return Iterator(&_root);
	}

	auto last() {
		if (begin() != end()) {
			return --end();
		}
		else {
			return end();
		}
	}

	auto rbegin() {
		return std::reverse_iterator(end());
	}

	auto rend() {
		return std::reverse_iterator(begin());
	}

	bool empty() const {
		return !_root.has_next();
	}

	void erase(T& item) {
		erase(get_item_node(item));
	}

	void insert(T& item, Iterator& target) {
		LinkedNode<T>& item_node = get_item_node(item);
		LinkedNode<T>& target_node = *target.node;
		item_node.unlink();
		item_node.previous = target_node.previous;
		item_node._value = target_node.previous->_value;
		target_node.previous->_value = &item;
		target_node.previous = &item_node;
	}

	void insert(T& item) {
		insert(item, end());
	}
private:
	LinkedNode<T>& get_item_node(T& item) {
		return *(LinkedNode<T>*)((char*)&item + _node_offset);
	}

	void erase(LinkedNode<T>& node) {
		node.unlink();
		node.clear();
	}

	void steal_content(TSLinkedList&& other) {
		_node_offset = other._node_offset;
		if (other.empty()) {
			_root.loop();
		}
		else {
			_root = other._root;
			_root.previous->_value = (T*)~(uintptr_t)&_root;
			_root.next(_node_offset).previous = &_root;
			other._root.loop();
		}
	}
};
