#pragma once

#include "wc3/agent.hpp"
#include "wc3/data_store.hpp"
#include "wc3/net_session.hpp"

const int FIRST_NET_COMMAND_EVENT = 0xA0000;

enum NetCommand: uint8_t {
	NET_COMMAND_PAUSE = 0x1,
	NET_COMMAND_UNPAUSE = 0x2,
	NET_COMMAND_GAME_SPEED = 0x3,
	NET_COMMAND_GAME_SPEED_INCREMENT = 0x4,
	NET_COMMAND_GAME_SPEED_DECREMENT = 0x5,
	NET_COMMAND_SAVE = 0x6,
	NET_COMMAND_SAVE_DONE = 0x7,
	NET_COMMAND_UNIT_ORDER_BASIC = 0x10,
	NET_COMMAND_UNIT_ORDER_TARGET_POINT = 0x11,
	NET_COMMAND_UNIT_ORDER_TARGET_IMAGE = 0x12,
	NET_COMMAND_UNIT_ORDER_TARGET_IMAGE2 = 0x13,
	NET_COMMAND_UNIT_ORDER_TARGET_IMAGE_FOGGED = 0x14,
	NET_COMMAND_UNIT_ORDER_TARGET_IMAGE_FOGGED2 = 0x15,
	NET_COMMAND_UNIT_SELECTION_MODIFY = 0x16,
	NET_COMMAND_UNIT_DEFINE_CONTROL_GROUP = 0x17,
	NET_COMMAND_UNIT_SELECT_CONTROL_GROUP = 0x18,
	NET_COMMAND_UNIT_SELECT_SUB_GROUP = 0x19,
	NET_COMMAND_UNIT_REFRESH_SUB_GROUP = 0x1A,
	NET_COMMAND_UNIT_SELECTION_EVENT = 0x1B,
	NET_COMMAND_SELECTABLE_SELECTION_MODIFY = 0x1C,
	NET_COMMAND_REVIVE_CANCEL = 0x1D,
	NET_COMMAND_TRAIN_CANCEL = 0x1E,
	NET_COMMAND_CHEAT_COOLDOWN = 0x20,
	NET_COMMAND_CHEAT_CRITTER = 0x21,
	NET_COMMAND_CHEAT_DEFEAT = 0x22,
	NET_COMMAND_CHEAT_FAST_BUILD = 0x23,
	NET_COMMAND_CHEAT_FAST_DEATH = 0x24,
	NET_COMMAND_CHEAT_FOOD = 0x25,
	NET_COMMAND_CHEAT_GOD = 0x26,
	NET_COMMAND_CHEAT_GOLD = 0x27,
	NET_COMMAND_CHEAT_LUMBER = 0x28,
	NET_COMMAND_CHEAT_MANA = 0x29,
	NET_COMMAND_CHEAT_NO_DEFEAT = 0x2A,
	NET_COMMAND_CHEAT_NO_VICTORY = 0x2B,
	NET_COMMAND_CHEAT_RESEARCH = 0x2C,
	NET_COMMAND_CHEAT_RESOURCES = 0x2D,
	NET_COMMAND_CHEAT_SET_TOD = 0x2E,
	NET_COMMAND_CHEAT_SHOW_MAP = 0x2F,
	NET_COMMAND_CHEAT_TECH_TREE = 0x30,
	NET_COMMAND_CHEAT_UPGRADE = 0x31,
	NET_COMMAND_CHEAT_VICTORY = 0x32,
	NET_COMMAND_TEAM_CHANGE_ALLIANCE = 0x50,
	NET_COMMAND_TEAM_TRADE_RESOURCES = 0x51,
	NET_COMMAND_PLAYER_CHAT_MATCH = 0x60,
	NET_COMMAND_PLAYER_END_CINEMATIC = 0x61,
	NET_COMMAND_RESUME_TRIGGER_EXEC = 0x62,
	NET_COMMAND_TRIGGER_SYNC_READY = 0x63,
	NET_COMMAND_TRACKABLE_HIT = 0x64,
	NET_COMMAND_TRACKABLE_TRACK = 0x65,
	NET_COMMAND_SHOW_SKILL = 0x66,
	NET_COMMAND_SHOW_SUBMENU = 0x67,
	NET_COMMAND_ALLY_PING = 0x68,
	NET_COMMAND_DIALOG_BUTTON_CLICK = 0x69,
	NET_COMMAND_DIALOG_CLICK = 0x6A,
	NET_COMMAND_SYNC_STORE_INTEGER = 0x6B,
	NET_COMMAND_SYNC_STORE_REAL = 0x6C,
	NET_COMMAND_SYNC_STORE_BOOLEAN = 0x6D,
	NET_COMMAND_SYNC_STORE_UNIT = 0x6E,
	NET_COMMAND_SYNC_STORE_STRING = 0x6F,
	NET_COMMAND_SYNC_CLEAR_INTEGER = 0x70,
	NET_COMMAND_SYNC_CLEAR_REAL = 0x71,
	NET_COMMAND_SYNC_CLEAR_BOOLEAN = 0x72,
	NET_COMMAND_SYNC_CLEAR_UNIT = 0x73,
	NET_COMMAND_SYNC_CLEAR_STRING = 0x74,
	NET_COMMAND_ARROW_KEY = 0x75,
	NET_COMMAND_REPLAY_DATA_UNCOMPRESSED = 0x81,
	NET_COMMAND_REPLAY_DATA_COMPRESSED = 0x82,
	NET_COMMAND_REPLAY_DATA_DONE = 0x83,
	NET_COMMAND_REPLAY_VIEW_SPEED_INCREMENT = 0x84,
	NET_COMMAND_REPLAY_VIEW_SPEED_DECREMENT = 0x85,
};

class CNetCommandBase: public CNetEventBase {
public:
	uint8_t command_id;
	uint8_t player_id;
};

class CNetCommandSave: public CNetCommandBase {
public:
	char filename[260];
};

struct CNetCommandUnitOrderBasic: public CNetCommandBase {
public:
	unsigned short flags;
	int order_id;
	NetTag item_tag;
};

CDataStore& operator << (CDataStore& store, const CNetCommandUnitOrderBasic& command);

class CNetCommandUnitOrderTargetPoint: public CNetCommandUnitOrderBasic {
public:
	float x;
	float y;
};

CDataStore& operator << (CDataStore& store, const CNetCommandUnitOrderTargetPoint& command);

class CNetCommandUnitOrderTargetImage: public CNetCommandUnitOrderTargetPoint {
public:
	NetTag target_tag;
};

CDataStore& operator << (CDataStore& store, const CNetCommandUnitOrderTargetImage& command);

class CNetCommandUnitOrderTargetImage2: public CNetCommandUnitOrderTargetImage {
public:
	NetTag target_item_tag;
};

CDataStore& operator << (CDataStore& store, const CNetCommandUnitOrderTargetImage2& command);

class CNetCommandUnitOrderTargetImageFogged: public CNetCommandUnitOrderTargetPoint {
public:
	int target_type_id;
	int flags2;
	int unk;
	int target_owner_id;
	float fogged_target_x;
	float fogged_target_y;
};

CDataStore& operator << (CDataStore& store, const CNetCommandUnitOrderTargetImageFogged& command);

class CNetCommandUnitOrderTargetImageFogged2: public CNetCommandUnitOrderTargetImageFogged {
public:
	NetTag fogged_item_tag;
};

CDataStore& operator << (CDataStore& store, const CNetCommandUnitOrderTargetImageFogged2& command);
