#include "wc3/command_bar.hpp"

#include "storm.h"
#include "wc3/game_base.hpp"

void CSimpleButton::click(MOUSE_BUTTON_FLAGS mouse_button) {
	static_cast<VMT*>(this->_vtable)->click(this, mouse_button);
}

CCommandButton* CCommandBar::get_button(int x, int y) {
	if (y < rows.size()) {
		Row row = rows[y];
		if (x < row.size()) {
			return static_cast<CCommandButton*>(row[x]);
		}
	}

	return nullptr;
}

void CCommandBar::update() {
	CCommandBar_Update(this);
}

void CCommandBar::add_command(CCommandButtonData& data) {
	CCommandBar_AddCommand(this, &data);
}

static uintptr_t CCommandBar_AddCommand_offset = get_offset(
	"CCommandBar::AddCommand",
	{
		{GameVersion::v1_26a, 0x36D3E0},
		{GameVersion::v1_27a, 0x38CCB0},
	}
);
TCCommandBar_AddCommand CCommandBar_AddCommand = (TCCommandBar_AddCommand)(get_game_base() + CCommandBar_AddCommand_offset);

static uintptr_t CCommandBar_Update_offset = get_offset(
	"CCommandBar::Update",
	{
		{GameVersion::v1_26a, 0x372F50},
		{GameVersion::v1_27a, 0x3A9530},
	}
);
TCCommandBar_Update CCommandBar_Update = (TCCommandBar_Update)(get_game_base() + CCommandBar_Update_offset);

int get_command_index(const CCommandButtonData& data, MOUSE_BUTTON_FLAGS for_button) {
	for (int i = 0; i < std::size(data.order_flags); i++) {
		switch (for_button) {
			case LEFT_MOUSE_BUTTON:
				if (data.order_flags[i]
					&& !(data.order_flags[i] & 0x40)
					&& !(data.order_flags[i] & 0x2000)
				) {
					return i;
				};
				break;
			case RIGHT_MOUSE_BUTTON:
				if (!(data.order_flags[i] & 0x40)
					&& (data.order_flags[i] & 0x2000)
				) {
					return i;
				}
			default:
				return -1;
		}
	}

	return -1;
}

bool CCommandBar::close_submenu() {
	auto last = submenu_items.last();
	if (last != submenu_items.end()) {
		SubMenuStackItem* submenu = last;
		submenu_items.erase(*last);
		SMemFree(submenu, __FILE__, __LINE__, 0);
		update();
		return true;
	}

	return false;
}
