#include "wc3/data_recycler.hpp"

#include <winnt.h>
#include "storm.h"

CDataRecycler::_Node* CDataRecycler::_get_node() {
	while (true) {
		_Node* node = _empty_nodes.pop();
		if (node) {
			return node;
		}

		_Cluster* cluster = (_Cluster*)SMemAlloc(
			sizeof(_Cluster) + sizeof(_Node) * _cluster_size,
			__FILE_NAME__,
			__LINE__,
			0
		);
		_clusters.push(cluster, 0);

		if (_cluster_size > 1) {
			for (int i = 0; i < _cluster_size - 1; i++) {
				cluster->nodes[i].next = &cluster->nodes[i + 1];
			}
		}

		_nodes.push(cluster->nodes, sizeof(_Node) * (_cluster_size - 1));
	}
}

bool CDataRecycler::get_data(void*& buffer, size_t& size) {
	_Node* node = _nodes.pop();
	if (node) {
		buffer = node->buffer;
		size = node->size;

		InterlockedIncrement(reinterpret_cast<volatile LONG*>(&_cache_slots_count));
		_empty_nodes.push(node, 0);

		return true;
	}

	return false;
}

void CDataRecycler::put_data(void* buffer, size_t size) {
	bool cache_limit_exceeded = InterlockedDecrement((volatile LONG*)&_cache_slots_count) < 0;
	if (!cache_limit_exceeded) {
		_Node* node = _get_node();
		node->buffer = buffer;
		node->size = size;
		_nodes.push(node, 0);
	}
	else {
		InterlockedIncrement((volatile LONG*)&_cache_slots_count);
		static_cast<_VMT*>(_vtable)->deallocate(this, buffer, nullptr, 0);
	}
}
