#include "wc3/data_store.hpp"
#include <algorithm>
#include <cstring>

bool CDataStore::_request_region(size_t block_position, size_t block_size) {
	return static_cast<VMT*>(vtable)->prepare_to_write(
		this,
		block_position,
		block_size,
		&data,
		&current_region_start,
		&capacity,
		__FILE_NAME__,
		__LINE__
	);
}

void CDataStore::clear() {
	static_cast<VMT*>(vtable)->clear(this);
}

bool CDataStore::write(const void* buffer, size_t count) {
	size_t bytes_left = count;

	while (bytes_left) {
		size_t chunk_size = bytes_left;

		if (position < current_region_start
			|| (position + chunk_size) > (current_region_start + capacity)
		) {
			if (!_request_region(position, chunk_size)) {
				// Class `CDataStoreChunked` has a limit on the maximum block size that can be requested.
				size_t minimum_size = std::min(bytes_left, std::max(1u, capacity));
				if (_request_region(position, minimum_size)) {
					size_t available_size = capacity - (position - current_region_start);
					chunk_size = std::min(bytes_left, available_size);
				}
				else {
					return false;
				}
			}
		}

		size_t total_bytes_written = count - bytes_left;
		memcpy(
			(char*)data + (position - current_region_start),
			(char*)buffer + total_bytes_written,
			chunk_size
		);

		bytes_left -= chunk_size;
	}

	return true;
}
